from datetime import datetime
import time
import os.path
import yaml
import logging
import sys
import threading
from progress.bar import IncrementalBar  # to show beautiful loading bar on the screen during sampling

# --------------------------------------------------
# I2C
# --------------------------------------------------
from smbus2 import SMBus
from sys import exit

# emplacement variable
bus = SMBus(1)

# attributed canals and associated emplacements variable
address = 0b1110110

# --------------------------------------------------------
# YAML SETTINGS
# --------------------------------------------------------

# Get current directory
current_working_directory = str(os.getcwd())

with open(current_working_directory + '/seacanairy_settings.yaml') as file:
    settings = yaml.safe_load(file)
    file.close()

store_debug_messages = settings['AFE Board']['Store debug messages (important increase of logs)']

project_name = settings['Seacanairy settings']['Sampling session name']


# --------------------------------------------------------
# LOGGING SETTINGS
# --------------------------------------------------------
# all the settings and other code for the logging
# logging = tak a trace of some messages in a file to be reviewed afterward (check for errors fe)


def set_logger(message_level, log_file):
    # set up logging to file
    logging.basicConfig(level=message_level,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%d-%m %H:%M',
                        filename=log_file,
                        filemode='a')

    logger = logging.getLogger('AFE Board')  # name of the logger
    # all further logging must be called by logger.'level' and not logging.'level'
    # if not, the logging will be displayed as 'ROOT' and NOT 'OPC-N3'
    return logger


if __name__ == '__main__':  # if you run this code directly ($ python3 CO2.py)
    message_level = logging.DEBUG  # show ALL the logging messages
    # Create a file to store the log if it doesn't exist
    log_file = current_working_directory + "/log/Alphasense_board-debugging.log"
    if not os.path.isfile(log_file):
        os.mknod(log_file)
    print("Alphasense Board DEBUG messages will be shown and stored in '" + str(log_file) + "'")
    logger = set_logger(message_level, log_file)
    # define a Handler which writes INFO messages or higher to the sys.stderr/display
    console = logging.StreamHandler()
    console.setLevel(message_level)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger().addHandler(console)

else:  # if this file is considered as a library (if you execute 'seacanairy.py' for example)
    # it will only print and store INFO messages and above in the corresponding log_file
    if store_debug_messages:
        message_level = logging.DEBUG
    else:
        message_level = logging.INFO
    log_file = '/home/pi/seacanairy_project/log/' + project_name + '-log.log'  # complete location needed on the RPI
    # no need to add a handler, because there is already one in seacanairy.py
    logger = set_logger(message_level, log_file)

# all further logging must be called by logger.'level' and not logging.'level'
# if not, the logging will be displayed as 'ROOT' and NOT 'GPS'

# --------------------------------------------------
# ADC channels
# --------------------------------------------------

# Channel Address - Single channel use
# See LTC2497 data sheet, Table 3, Channel Selection.
# All channels are uncommented - comment out the channels you do not plan to use.

channel0 = 0xB0
channel1 = 0xB8
channel2 = 0xB1
channel3 = 0xB9
channel4 = 0xB2
channel5 = 0xBA
channel6 = 0xB3
channel7 = 0xBB
channel8 = 0xB4
channel9 = 0xBC
channel10 = 0xB5
channel11 = 0xBD
channel12 = 0xB6
channel13 = 0xBE
channel14 = 0xB7
channel15 = 0xBF

# reference voltage of the ADC
vref = 5

# To calculate the voltage, the number read in is 3 bytes. The first bit is ignored.
# Max reading is 2^23 or 8,388,608
max_reading = 8388608.0

# lange = number of bytes to read. A minimum of 3 bytes are read in.
# In this sample we read in 6 bytes, ignoring the last three bytes
# zeit = tells how frequently you want the readings to be read from the ADC.
# Define the time to sleep between the readings.
# tiempo = shows how frequently each channel is read in over the I2C bus.
# Best to use timepo between each successive readings.

lange = 0x06  # number of bytes to read in the block
zeit = 15  # number of seconds to sleep between each measurement
sleep = 0.2  # number of seconds to sleep between each channel reading

# has to be more than 0.2 (seconds)

# ----------------------------------------
# CALIBRATION INFORMATION
# ----------------------------------------

# Temperature
with open(current_working_directory + '/AFE_calibration/temperature_calib.yaml') as file:
    temp_calib = yaml.safe_load(file)
    file.close()

# NO2
with open(current_working_directory + '/AFE_calibration/NO2_calib.yaml') as file:
    NO2_calib = yaml.safe_load(file)
    file.close()

# SO2
with open(current_working_directory + '/AFE_calibration/SO2_calib.yaml') as file:
    SO2_calib = yaml.safe_load(file)
    file.close()

# OX
with open(current_working_directory + '/AFE_calibration/OX_calib.yaml') as file:
    OX_calib = yaml.safe_load(file)
    file.close()

# CO
with open(current_working_directory + '/AFE_calibration/CO_calib.yaml') as file:
    CO_calib = yaml.safe_load(file)
    file.close()


def getADCreading(adc_address, adc_channel):
    """
    Read tension from the ADC on a certain channel

    :param adc_address: slave i2c address
    :param adc_channel: channel where to read tension
    :return: tension between channel and ground (volts)
    """

    attempts = 0

    while attempts < 4:

        try:
            bus.write_byte(adc_address, adc_channel)
            # print("Reading tension...                                         ", end='\r')
            time.sleep(sleep)
            reading = bus.read_i2c_block_data(adc_address, adc_channel, lange)
            # ----------- Start conversion for the Channel Data ----------
            valor = ((((reading[0] & 0x3F)) << 16)) + ((reading[1] << 8)) + (((reading[2] & 0xE0)))
            # add a debug function
            # debug(print("Valor is 0x%x" % valor))

            # ----------- End of conversion of the Channel ----------
            volts = round(valor * vref / max_reading, 7)
            # Rounding to 7 decimals because ADC accuracy is 3.9 microvolt
            # print("Reading tension...", volts, "V", end='\r')

            if (reading[0] & 0b11000000) == 0b11000000:
                logger.error(
                    "Input voltage is either open or more than " + str(vref) + "Volts.")
                logger.warning("The reading may not be correct. Value read is " + str(volts) + " mV")

            # time.sleep(sleep)  # be sure to have some time laps between two I2C reading/writing # i2c don't care!
            return volts

        except:
            if attempts >= 3:
                logger.critical("i2c transmission failed 3 consecutive times(" + str(sys.exc_info())
                                + "), skipping i2c reading")
                return False  # indicate clearly that system has failed

            logger.error("Error in the i2c transmission (" + str(sys.exc_info())
                         + "), trying again... (" + str(attempts) + "/3)")
            attempts += 1  # increment of reading_trials
            time.sleep(1)  # if transmission fails, wait a bit to try again (sensor is maybe busy)

    return False


# ====================================================================================


reading_multiplier = 1000  # multiplication of the value given by the rpi


def read_temp():
    """
    Measure tension of the temperature sensor
    (Note that sensor is not located in the gas hood.)

    :return: Dictionary containing tension in milli volts {'temperature raw'}
    """

    volts = getADCreading(address, channel5)
    if volts is not False:
        tempv = round(reading_multiplier * volts, 5)
        logger.debug("Tension from temperature sensor (AFE board) is " + str(tempv) + " mV")
        time.sleep(sleep)

        temp_to_return = {
            "temperature raw": tempv,
            "temperature": "-"
        }
    else:
        logger.critical("Failed to read temperature")
        temp_to_return = {
            "temperature raw": "error",
            "temperature": "error"
        }

    return temp_to_return


def read_NO2():
    """
    Measure tension of NO2 main and auxiliary electrodes

    :return: Dictionary containing tensions in milli volts {'NO2 main', 'NO2 aux'}
    """
    volts = getADCreading(address, channel8)
    if volts is not False:
        NO2v_main = round(reading_multiplier * volts, 5)
        logger.debug("Tension from NO2 sensor (main) is " + str(NO2v_main) + " mV")
        time.sleep(sleep)
        NO2v_aux = round(reading_multiplier * getADCreading(address, channel4), 5)
        logger.debug("Tension from NO2 sensor (aux) is " + str(NO2v_aux) + " mV")
        time.sleep(sleep)

        NO2_to_return = {
            "NO2 main": NO2v_main,
            "NO2 aux": NO2v_aux,
            "NO2 ppb": "-"
        }
    else:
        logger.critical("Failed to read NO2 sensor")
        NO2_to_return = {
            "NO2 main": "error",
            "NO2 aux": "error",
            "NO2 ppb": "error"
        }

    return NO2_to_return


def read_OX():
    """
    Measure tension of OX main and auxiliary electrodes

    :return: Dictionary containing tensions in milli volts {'OX main', 'OX aux'}
    """
    volts = getADCreading(address, channel7)
    if volts is not False:
        Oxv_main = round(reading_multiplier * volts, 5)
        logger.debug("Tension from Ox sensor (main) is " + str(Oxv_main) + " mV")
        time.sleep(sleep)
        Oxv_aux = round(reading_multiplier * getADCreading(address, channel3), 5)
        logger.debug("Tension from Ox sensor (aux) is " + str(Oxv_aux) + " mV")
        time.sleep(sleep)

        OX_to_return = {
            "OX main": Oxv_main,
            "OX aux": Oxv_aux,
            "OX ppb": "-"
        }

    else:
        logger.critical("Failed to read OX")
        OX_to_return = {
            "OX main": "error",
            "OX aux": "error",
            "OX ppb": "error"
        }

    return OX_to_return


def read_SO2():
    """
    Measure tension of SO2 main and auxiliary electrodes

    :return: Dictionary containing tensions in milli volts {'SO2 main', 'SO2 aux'}
    """
    volts = getADCreading(address, channel6)
    if volts is not False:
        SO2v_main = round(reading_multiplier * volts, 5)
        logger.debug("Tension from SO2 sensor (main) is " + str(SO2v_main) + " mV")
        time.sleep(sleep)
        SO2v_aux = round(reading_multiplier * getADCreading(address, channel2), 5)
        logger.debug("Tension from SO2 sensor (aux) is " + str(SO2v_aux) + " mV")
        time.sleep(sleep)

        SO2_to_return = {
            "SO2 main": SO2v_main,
            "SO2 aux": SO2v_aux,
            "SO2 ppb": "-"
        }

    else:
        logger.critical("Failed to read SO2")

        SO2_to_return = {
            "SO2 main": "error",
            "SO2 aux": "error",
            "SO2 ppb": "error"
        }

    return SO2_to_return


def read_CO():
    """
    Measure tension of CO main and auxiliary electrodes

    :return: Dictionary containing tensions in milli volts {'CO main', 'CO aux'}
    """
    volts = getADCreading(address, channel0)
    if volts is not False:
        COv_main = round(reading_multiplier * volts, 5)
        time.sleep(sleep)
        logger.debug("Tension from CO sensor (main) is " + str(COv_main) + " mV")
        COv_aux = round(reading_multiplier * getADCreading(address, channel1), 5)
        logger.debug("Tension from CO sensor (aux) is " + str(COv_aux) + " mV")
        time.sleep(sleep)

        CO2_to_return = {
            "CO main": COv_main,
            "CO aux": COv_aux,
            "CO ppb": "-"
        }
    else:
        logger.critical("Failed to read CO")

        CO2_to_return = {
            "CO main": "error",
            "CO aux": "error",
            "CO ppb": "error"
        }

    return CO2_to_return


def calibrate_temperature(main):
    """
    Apply calibration to the temperature measurement
    :param main: electric tension from the temperature sensor
    :return: temperature in °C
    """
    temperature = ((main - temp_calib["Vkal"]) / temp_calib["Thermal sensitivity"]) \
                  + temp_calib["Tkal"]

    temperature = round(temperature, 2)

    return temperature


def calibrate_NO2(main, aux):
    """
    Apply calibration to the NO2 measurements
    :param main: electrical tension from the main electrode
    :param aux: electrical tension from the auxiliary electrode
    :return: NO2 concentration (ppb)
    """
    # Algorithm 1
    NO2_ppb = (
                      (

                              (main - NO2_calib["WE0_e"]) -
                              (
                                      NO2_calib["nt"] * (aux - NO2_calib["AE0_e"])
                              )
                      )
                      / NO2_calib["WE_SENS"]
              ) + NO2_calib["C"]

    NO2_ppb = round(NO2_ppb, 1)

    return NO2_ppb


def calibrate_OX(main, aux, NO2_ppb):
    """
    Apply calibration to the OX measurements
    :param main: electrical tension from the main electrode
    :param aux: electrical tension from the auxiliary electrode
    :param NO2_ppb: NO2 concentration from the other sensor
    :return: OX concentration (ppb)
    """
    # Algorithm 3
    O3_ppb = (
                     (
                             (main - OX_calib["WE0_e"] - (NO2_ppb * OX_calib["WE_SENS_NO2"])
                              ) -
                             (OX_calib["WE0_s"] - OX_calib["AE0_s"]) -
                             (
                                     OX_calib["nt"] *
                                     (aux - OX_calib["AE0_e"])
                             )
                     ) / OX_calib["WE_SENS"]
             ) \
             + OX_calib["C"]

    O3_ppb = round(O3_ppb, 1)

    return O3_ppb


def calibrate_SO2(main, aux):
    """
    Apply calibration to the SO2 measurements
    :param main: electrical tension from the main electrode
    :param aux: electrical tension from the auxiliary electrode
    :return: SO2 concentration (ppb)
    """
    # Algorithm 4
    SO2_ppb = -1 * \
              (
                      (
                              (main - SO2_calib["WE0_e"]) - SO2_calib["WE0_s"] - SO2_calib["nt"]
                      )
                      / SO2_calib["WE_SENS"]) \
              + SO2_calib["C"]

    SO2_ppb = round(SO2_ppb, 1)

    return SO2_ppb


def calibrate_CO(main, aux):
    """
    Apply calibration to the CO measurements
    :param main: electrical tension from the main electrode
    :param aux: electrical tension from the auxiliary electrode
    :return: CO concentration (ppb)
    """
    # Algorithm 1
    CO_ppb = (
                     (
                             (main - CO_calib["WE0_e"]) -
                             (
                                     CO_calib["nt"] *
                                     (aux - CO_calib["AE0_e"])
                             )
                     )
                     / CO_calib["WE_SENS"]) \
             + CO_calib["C"]

    CO_ppb = round(CO_ppb, 1)

    return CO_ppb


def calibrate_all(data):
    """
    Apply calibration to all available items
    :param data: dictionary containing values to calibrate
    :return: dictionnary with calibrated items
    """

    if "temperature raw" in data:
        data["temperature"] = calibrate_temperature(data["temperature raw"])

    if "NO2 main" in data:
        data["NO2 ppb"] = calibrate_NO2(data["NO2 main"], data["NO2 aux"])
        if "OX main" in data:
            data["OX ppb"] = calibrate_OX(data["OX main"], data["OX aux"], data["NO2 ppb"])

    if "SO2 main" in data:
        data["SO2 ppb"] = calibrate_SO2(data["SO2 main"], data["SO2 aux"])

    if "CO main" in data:
        data["CO ppb"] = calibrate_CO(data["CO main"], data["CO aux"])

    return data


def get_data():
    """
    Get all available data from the 4-AFE Alphasense Board (one signe instantaneous reading)

    :return: dictionary{'NO2 ppb', NO2 main', 'NO2 aux', 'OX ppb', 'OX main', 'OX aux',
                'SO2 ppb', 'SO2 main', 'SO2 aux', 'CO ppb', 'CO main', 'CO aux',
                'temperature', 'temperature raw'}
    """

    data = {}

    data.update(read_NO2())
    data.update(read_OX())
    data.update(read_SO2())
    data.update(read_CO())
    data.update(read_temp())
    data = calibrate_all(data)

    return data


def print_measurements(data):
    """
    Print all measurement data on the screen
    :param data: dictionary containing all the data
    :return: nothing
    """
    print("\t\t   ppb \t\t( main (mV), aux (mV) )")
    print("NO2:\t\t", data["NO2 ppb"], "\t\t(", data["NO2 main"], ", ", data["NO2 aux"], ")")
    print("OX:\t\t", data["OX ppb"], "\t\t(", data["OX main"], ", ", data["OX aux"], ")")
    print("SO2:\t\t", data["SO2 ppb"], "\t\t(", data["SO2 main"], ", ", data["SO2 aux"], ")")
    print("CO:\t\t", data["CO ppb"], "\t\t(", data["CO main"], ", ", data["CO aux"], ")")
    print("Temperature:\t", data["temperature"], "\t\t(", data["temperature raw"], ")")

    return


def start_averaged_data(number_of_measurements, delay = 0, display=True):
    """
    Perform multiple readings and makes an average
    Run get_averaged_data() once thread is finished to get the data
    Improved for threading application (no display prints)

    :param number_of_measurements: number of measurement to average, each single measurement taking around 2 seconds
    :return: Dictionary{'NO2 main', 'NO2 aux', 'OX main', 'OX aux',
                'SO2 main', 'SO2 aux', 'CO main', 'CO aux',
                'temperature raw'}
    """
    global thread_data

    thread_data = {}
    if delay:
        time.sleep(delay)

    if display:
        bar = IncrementalBar("Reading tensions", max=(5 * number_of_measurements))

    logger.debug("Starting averaged data reading")
    NO2_main = []
    NO2_aux = []

    OX_main = []
    OX_aux = []

    SO2_main = []
    SO2_aux = []

    CO_main = []
    CO_aux = []

    temperature_main = []

    for _ in range(number_of_measurements):
        NO2 = read_NO2()
        if display: bar.next()
        NO2_main += [NO2['NO2 main']]
        NO2_aux += [NO2['NO2 aux']]

    thread_data.update({"NO2 main min": min(NO2_main)})
    thread_data.update({"NO2 main max": max(NO2_main)})
    thread_data.update({"NO2 aux min": min(NO2_aux)})
    thread_data.update({"NO2 aux max": max(NO2_aux)})

    sum = 0
    for i in range(len(NO2_main)):
        sum += NO2_main[i]
    NO2_main = sum / len(NO2_main)

    sum = 0
    for i in range(len(NO2_aux)):
        sum += NO2_aux[i]
    NO2_aux = sum / len(NO2_aux)

    for _ in range(number_of_measurements):
        OX = read_OX()
        if display: bar.next()
        OX_main += [OX['OX main']]
        OX_aux += [OX['OX aux']]

    thread_data.update({"OX main min": min(OX_main)})
    thread_data.update({"OX main max": max(OX_main)})

    thread_data.update({"OX aux min": min(OX_aux)})
    thread_data.update({"OX aux max": max(OX_aux)})

    sum = 0
    for i in range(len(OX_main)):
        sum += OX_main[i]
    OX_main = sum / len(OX_main)

    sum = 0
    for i in range(len(OX_aux)):
        sum += OX_aux[i]
    OX_aux = sum / len(OX_aux)

    for _ in range(number_of_measurements):
        SO2 = read_SO2()
        if display: bar.next()
        SO2_main += [SO2['SO2 main']]
        SO2_aux += [SO2['SO2 aux']]

    thread_data.update({"SO2 main min": min(SO2_main)})
    thread_data.update({"SO2 main max": max(SO2_main)})
    thread_data.update({"SO2 aux min": min(SO2_aux)})
    thread_data.update({"SO2 aux max": max(SO2_aux)})

    sum = 0
    for i in range(len(SO2_main)):
        sum += SO2_main[i]
    SO2_main = sum / len(SO2_main)

    sum = 0
    for i in range(len(SO2_aux)):
        sum += SO2_aux[i]
    SO2_aux = sum / len(SO2_aux)

    for _ in range(number_of_measurements):
        CO = read_CO()
        if display: bar.next()
        CO_main += [CO['CO main']]
        CO_aux += [CO['CO aux']]

    thread_data.update({"CO main min": min(CO_main)})
    thread_data.update({"CO main max": max(CO_main)})
    thread_data.update({"CO aux min": min(CO_aux)})
    thread_data.update({"CO aux max": max(CO_aux)})

    sum = 0
    for i in range(len(CO_main)):
        sum += CO_main[i]
    CO_main = sum / len(CO_main)

    sum = 0
    for i in range(len(CO_aux)):
        sum += CO_aux[i]
    CO_aux = sum / len(CO_aux)

    for _ in range(number_of_measurements):
        temp = read_temp()
        if display: bar.next()
        temperature_main += [temp['temperature raw']]

    thread_data.update({"temperature min": min(temperature_main)})
    thread_data.update({"temperature max": max(temperature_main)})

    sum = 0

    for i in range(len(temperature_main)):
        sum += temperature_main[i]
    temperature = sum / len(temperature_main)

    thread_data.update({
        'NO2 main': NO2_main,
        'NO2 aux': NO2_aux,
        'OX main': OX_main,
        'OX aux': OX_aux,
        'SO2 main': SO2_main,
        'SO2 aux': SO2_aux,
        'CO main': CO_main,
        'CO aux': CO_aux,
        'temperature raw': temperature
    })

    thread_data = calibrate_all(thread_data)

    logger.debug("Execute function 'get_averaged_data' to show data on screen")

    bar.finish()

    return thread_data


def start_background_average_measurement(number_of_measurements, delay=0):
    """
    Start a new thread to perform averaged reading in the background
    Run get_averaged_data() once thread is finished to get the data

    thread = threading.Thread(target=AFE.start_averaged_data, args=([number_of_measurements, delay]), daemon=True)
    in your own code is preferred

    :param number_of_measurements: number of measurement to average, each single measurement taking around 2 seconds
    :param delay: amount of time in between the start of the thread and the start of the sampling operation
    :return: nothing
    """

    print("Starting background AFE average reading...")
    x = threading.Thread(target=start_averaged_data, args=([number_of_measurements, delay]), daemon=True)
    x.start()
    logger.debug("Execute function 'get_averaged_data' to show data on screen")
    return


def get_averaged_data():
    """
    Read the data of the last start_averaged_data() (or start_background_average_measurement()) performed

    :return: dictionary{'NO2 ppb', NO2 main', 'NO2 aux', 'OX ppb', 'OX main', 'OX aux',
                'SO2 ppb', 'SO2 main', 'SO2 aux', 'CO ppb', 'CO main', 'CO aux',
                'temperature', 'temperature raw'}
    """
    global thread_data

    print_measurements(thread_data)

    return thread_data


# open the file where the data will be stored
if __name__ == "__main__":
    # Execute an execution test if the script is executed from there
    while True:
        get_data()
