import yaml
import logging
import sys
import mysql.connector
import os

# --------------------------------------------------------
# YAML SETTINGS
# --------------------------------------------------------

# Get current directory
current_working_directory = str(os.getcwd())

with open(current_working_directory + '/seacanairy_settings.yaml') as file:
    settings = yaml.safe_load(file)
    file.close()

store_debug_messages = settings['MySQL database settings']['Store debug messages (important increase of logs)']

project_name = settings['Seacanairy settings']['Sampling session name']

host = settings['MySQL database settings']['Host']
user = settings['MySQL database settings']['User']
password = settings['MySQL database settings']['Password']
db_name = settings['MySQL database settings']['Database name']
table_name = project_name


# --------------------------------------------------------
# LOGGING SETTINGS
# --------------------------------------------------------
# all the settings and other code for the logging
# logging = tak a trace of some messages in a file to be reviewed afterward (check for errors fe)


def set_logger(message_level, log_file):
    # set up logging to file
    logging.basicConfig(level=message_level,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%d-%m %H:%M',
                        filename=log_file,
                        filemode='a')

    logger = logging.getLogger('MySQL')  # name of the logger
    # all further logging must be called by logger.'level' and not logging.'level'
    # if not, the logging will be displayed as 'ROOT' and NOT 'OPC-N3'
    return logger


if __name__ == '__main__':  # if you run this code directly ($ python3 CO2.py)
    message_level = logging.DEBUG  # show ALL the logging messages
    # Create a file to store the log if it doesn't exist
    log_file = current_working_directory + "/log/mysql-debugging.log"
    if not os.path.isfile(log_file):
        os.mknod(log_file)
    print("MySQL DEBUG messages will be shown and stored in '" + str(log_file) + "'")
    logger = set_logger(message_level, log_file)
    # define a Handler which writes INFO messages or higher to the sys.stderr/display
    console = logging.StreamHandler()
    console.setLevel(message_level)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger().addHandler(console)

else:  # if this file is considered as a library (if you execute 'seacanairy.py' for example)
    # it will only print and store INFO messages and above in the corresponding log_file
    if store_debug_messages:
        message_level = logging.DEBUG
    else:
        message_level = logging.INFO
    log_file = '/home/pi/seacanairy_project/log/' + project_name + '-log.log'  # complete location needed on the RPI
    # no need to add a handler, because there is already one in seacanairy.py
    logger = set_logger(message_level, log_file)

# all further logging must be called by logger.'level' and not logging.'level'
# if not, the logging will be displayed as 'ROOT' and NOT 'MySQL'

# ---------------------------------------
# CODE
# ---------------------------------------

# Global variables
global connected
connected = False
table_created = False
global data_in_cache
data_in_cache = False
global db_line_count
db_line_count = False
global datafile_length
datafile_length = False

# Connect to the server


print("Connecting to MySQL database...", end='\r')


def connect(print_status=False):
    """
    Check connection and connect if connection is lost
    :param print_status: decide to show connection status or not
    :return: True (connected) or False (Not connected)
    """
    global mydb  # share variable through this code file
    global connected

    print("Checking MySQL connection status...", end='\r')

    if connected:  # if connection has already been established before
        return connected
    else:  # if first execution, or if connection has failed last time
        try:
            mydb = mysql.connector.connect(
                host=host,
                user=user,
                password=password,
                database=db_name)
            connected = mydb.is_connected()
            if connected:
                if print_status:
                    logger.info("Connected to database for online data storage")
                return connected
            else:
                if print_status:
                    logger.warning(
                        "No connection to database. Check internet status and database information in the settings")
                return connected
        except:
            if print_status:
                logger.error("Failed to connect to database (" + str(sys.exc_info()) + ")")
            connected = False
            return connected


def create_new_table(header_list, header_type):
    """
    Create a new table in the database
    :param header_list: list of all the column headers
    :return:
    """
    global connected
    global table_created
    header_name = str(header_list[0]) + " " + header_type[0]
    for i in range(1, len(header_list)):
        header_name += ", " + header_list[i] + " " + header_type[i]

    if connect():
        try:
            print("Creating database table '" + str(table_name) + "'...", end='\r')
            mycursor = mydb.cursor()
            mycursor.execute("CREATE TABLE IF NOT EXISTS `" + str(table_name) + "` (" + header_name + ")")
            logger.info("Table (already) created (" + str(table_name) + ")")
            table_created = True
            return True
        except:
            logger.error("Failed to create new table in the database (" + str(sys.exc_info()) + ")")
            connected = False
            return False


def number_of_lines_in_db():
    """
    Count the number of lines in the current table
    :return: number of lines
    """
    global connected
    try:
        mycursor = mydb.cursor()
        mycursor.execute("SELECT COUNT(*) FROM `" + str(table_name) + "`")
        result = mycursor.fetchone()
        print("There is already", result[0], "lines in db")
        return result[0]
    except:
        connected = False
        logger.error("Failed to count the number of lines in the database (" + str(sys.exc_info()) + ")")
        return 0


def upload_data(header_list, list_of_lines):
    """
    Append data to the table in the database
    :param header_list: List of the headers in which to write some data
    :param data: List containing the data to fill in those headers, in good order
    :return: True or False (success or failure)
    """
    global db_line_count
    # Convert header list in string
    # Necessary to create the MySQL function
    header_name = str(header_list[0])
    for i in range(1, len(header_list)):
        header_name += ", " + str(header_list[i])

    header_name = header_name.replace('"', '')  # no guillemet for header creation (see MySQL theory)

    # print("Header name is:", header_name)
    # print("Len header is:", len(header_list))
    # print("len of lines to send is:", len(list_of_lines))

    for i in range(len(list_of_lines)):
        # Create data string from list
        to_write = []
        for j in list_of_lines[i]:
            if isinstance(j, list):
                to_write += j
            else:
                to_write += [j]

        to_upload = '"' + to_write[0] + '"'  # the first one will always be the datetime, requires guillemets

        for j in range(1, len(to_write)):
            if to_write[j] == '-' or to_write[j] == 'error' or to_write[j] == '-\n' or to_write[j] == 'error\n' or \
                    to_write[j] == 'no fix':
                to_upload += ',NULL'
            else:
                try:
                    to_upload += ',' + str(float(to_write[j].replace('\n', '')))
                except:
                    to_upload += ', "' + str(to_write[j]).replace('\n', '') + '"'

        # print("To upload is:", to_upload)
        # print("Len to_upload is:", len(data))
        sql = "-"
        print("Sending", str(i) + '/' + str(len(list_of_lines)), " lines to MySQL database...", end='\r')

        if "0x00" in to_upload:  # if line in file is corrupted, skip this line and go ahead
            try:
                sql = "INSERT INTO `" + str(table_name) + "`"
                mycursor = mydb.cursor()
                mycursor.execute(sql)
                mydb.commit()
                db_line_count += 1  # one line sent, database is now one line bigger
            except:
                pass
            continue
        try:
            # print("INSERT INTO `" + str(table_name) + "` (" + header_name + ") VALUES (" + str(to_upload) + ")")
            mycursor = mydb.cursor()
            sql = "INSERT INTO `" + str(table_name) + "` (" + header_name + ") VALUES (" + str(
                to_upload) + ")"
            mycursor.execute(sql)
            mydb.commit()
            db_line_count += 1  # one line sent, database is now one line bigger
        except:
            logger.error("Failed to save data into the database (" + str(sys.exc_info()) + "; line was: ", str(sql),
                         ")")
            sql = "INSERT INTO `" + str(table_name) + "`"
            mycursor = mydb.cursor()
            mycursor.execute(sql)
            mydb.commit()
            db_line_count += 1  # one line sent, database is now one line bigger
    print("                                                                            ", end='\r')
    print(len(list_of_lines), "line(s) uploaded on the database")


def csv_file_length():
    """
    Get data file length
    :return: data file length
    """
    global datafile_length
    file = open(current_working_directory + "/" + str(project_name) + "/" + str(project_name) + "-data.csv", 'r')
    datafile_length = 0
    for line in file:
        if line != "\n":
            datafile_length += 1
    file.close()
    del file  # remove from memory
    return datafile_length - 1  # minus 1 for the column header line


def update(header_list, header_type, data_to_add):
    """
    Upload data on the database if possible
    :return:
    """
    global table_created
    global db_line_count
    global datafile_length
    global connected

    if not datafile_length:
        datafile_length = csv_file_length()
    # print("file length is:", datafile_length)

    if connect():  # check connection, and go ahead if ok
        if not table_created:
            try:
                table_created = create_new_table(header_list, header_type)
            except:
                print("Impossible to create table (" + str(sys.exc_info()) + ")")
                connected = False
                return  # impossible to store data if table doesn't exist
        if not db_line_count:
            db_line_count = number_of_lines_in_db()
        # print("db length is:", db_line_count)

        for i in range(1, len(data_to_add)):
            if type(data_to_add[i]) == str:
                # in case of any sensor error, it return either "-" or "error"
                # a float value could become a string, leading to database failure
                data_to_add[i].replace('-', 'NULL').replace('error', 'NULL').replace('no fix', 'NULL')
                # in MySQL, NULL means 'empty'

        if db_line_count == datafile_length:  # if the same amount of data in the db and in the csv file
            upload_data(header_list, [data_to_add])  # send only the newest data

        else:
            logger.info("Sending pending data to MySQL database")
            to_upload = []
            csv_file = open(
                current_working_directory + "/" + str(project_name) + "/" + str(project_name) + "-data.csv", 'r')
            csv_lines = csv_file.readlines()  # load all the lines in the memory
            csv_file.close()
            for x in range(db_line_count + 1, datafile_length + 1):
                to_upload += [csv_lines[x].split(',')]
                # for i in range(1, len(csv_lines)):
                #     if type(data_to_add[i]) == str:
                #         csv_lines[i].replace('-', 'NULL').replace('error', 'NULL')
            to_upload += [data_to_add]
            upload_data(header_list, to_upload)
            del csv_lines  # remove this huge variable from RAM
            del csv_file

    else:
        return

    datafile_length += 1
    # if this function is executed, it means that new data have been taken
    # therefore, we know the data file will have one line more
    # this way, we avoid loading each time the csv file
    # same increment after every database upload
