#! /home/pi/seacanairy_project/venv/bin/python3
"""
This code aims to make the whole Seacanairy works! It start the pump, read the date from the CO2 sensor,
the OPC-N3, the AFE board, the GPS, stop the pump, and store the whole in a file.
"""

# import all libraries
import time
from datetime import date, datetime, timedelta
import csv  # for storing data in file
from progress.bar import IncrementalBar  # to show beautiful loading bar on the screen during sampling
import os  # to be able to create new files/folders and see the current path
import yaml  # to read the settings stored in the 'seacanairy_settings.yaml' file
import logging  # to store the errors messages in a separate log file
import threading
import sys  # for error messages

# ---------------------------------------
# HELLO
# ---------------------------------------
print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
print("~~~~~~~~~~~~~~~~ SEACANAIRY ~~~~~~~~~~~~~~~~")
print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

# ---------------------------------------
# CHECK THE TIME
# ---------------------------------------
print("\n################# TIME #################")
print("Current date and time is:", datetime.now())
print("If time is not correct, close this window and click on 'Check Date and Time'")

# ---------------------------------------
# SETTINGS
# ---------------------------------------
# Import the settings from the Yaml file
# Store the different settings in dedicated variables to get it more easy

print("\n################ FILES #################")

current_working_directory = str(os.getcwd())  # returns the path where the python script is currently running

with open(current_working_directory + '/seacanairy_settings.yaml') as file:
    settings = yaml.safe_load(file)  # load all the yaml file in a dictionary
    file.close()  # always close the file after use

# Sampling period
sampling_period = settings['Seacanairy settings']['Sampling period']

# CO2 sampling period (CO2 sensor takes samples automatically)
CO2_sampling_period = int(sampling_period / settings['CO2 sensor']['Automatic sampling frequency ' \
                                                                   '(number of sample during the above sampling period)'])

# Amount of time required for the CO2 sensor to take the measurement
CO2_startup_delay = settings['CO2 sensor']['Amount of time required for the sensor to take the measurement']

# Name of the research (f-e 'Air measurement in my room')
project_name = settings['Seacanairy settings']['Sampling session name']

# Amount of time while the OPCN3 fan keep running without measurement to flush fresh air inside the casing
OPC_flushing_time = settings['OPC-N3 sensor']['Flushing time']

# Amount of time the OPCN3's laser takes PM sample
OPC_sampling_time = settings['OPC-N3 sensor']['Sampling time']

# OPCN3 Fan speed (0-100)
OPC_fan_speed = settings['OPC-N3 sensor']['Fan speed']

# Number of reading AFE must take and average
AFE_readings_averaged = settings['AFE Board']['Noise reduction - number of reading averaged']

# Let the user choose if he want to activate the following sensor or not
# Seen the problems encountered with GPS and OPCN3, could be good to disable the unnecessary sensors (GPS f-e)
# This does not shut down the sensor powering
CO2_activation = settings['CO2 sensor']['Activate this sensor']
OPCN3_activation = settings['OPC-N3 sensor']['Activate this sensor']
GPS_activation = settings['GPS']['Activate this sensor']
AFE_activation = settings['AFE Board']['Activate this sensor']
air_pump_activation = settings['Seacanairy settings']['Activate M&C air pump']
flow_sensor_activation = settings['Air flow sensor']['Activate this sensor']
mysql = settings['MySQL database settings']['Activate database upload']
air_pump_minimum_running_time = settings['Seacanairy settings']['Air pump minimum running time']
AFE_delay = settings['AFE Board']["Absorption time between air pump stop and reading"]

# -----------------------------------------
# CREATE FILES
# -----------------------------------------

# Create a directory to store everything if it doesn't exit
directory_path = current_working_directory + "/" + project_name
if not os.path.exists(directory_path):
    os.mkdir(directory_path)  # create the directory
    print("Created directory", directory_path)

# Create a file to store the log if it doesn't exist
log_file = directory_path + "/" + project_name + "-log.log"
if not os.path.isfile(log_file):
    os.mknod(log_file)  # create the file
    print("Created log file", log_file)
else:
    # Add a line to separate different executions
    file_logfile = open(log_file, "a")
    file_logfile.write("********************** NEW SESSION **********************\n")
    file_logfile.close()
# You can't go further in the program without creating the folder and the logger file

# -----------------------------------------
# IMPORT NECESSARY SENSORS
# -----------------------------------------
# Seen the above settings, import the necessary libraries
# Avoid starting module (i2c, SPI, UART) that will not be used

if CO2_activation:
    import CO2  # Library for the CO2 sensor

if OPCN3_activation:
    import OPCN3  # Library for the OPC-N3 sensor

if GPS_activation:
    import GPS  # Library for the GPS

if AFE_activation:
    import AFE

if air_pump_activation:
    import RPi.GPIO as GPIO  # to put GPIO high/low to switch the air pump relay on and off

if flow_sensor_activation:
    import flow as flow

if mysql:
    import database as db

# -----------------------------------------
# LOGGING
# -----------------------------------------
# Save the messages shown on the console in a dedicated file to understand the possible issues afterwards

message_level = logging.INFO

# set up logging to file
logging.basicConfig(level=message_level,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%d-%m %H:%M:%S',
                    filename=log_file,
                    filemode='a')
# define a Handler which writes INFO messages or higher to the sys.stderr/display
console = logging.StreamHandler()
console.setLevel(message_level)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger().addHandler(console)

logger = logging.getLogger('SEACANAIRY')

# Following logging messages must be called by logger.debug (...)

# -----------------------------------------
# FUNCTIONS
# -----------------------------------------
# Use the GPIO to turn on and off the air pump via relay

if air_pump_activation:
    GPIO.setwarnings(False)  # avoid messages concerning GPIO (GPIO is already low, blablabla)
    GPIO.setmode(GPIO.BCM)  # use the GPIO names (GPIO1...) instead of the processor pin name (BCM...)
    pump_gpio = 27
    GPIO.setup(pump_gpio, GPIO.OUT, initial=GPIO.LOW)


# -----------------------------------------
# FUNCTIONS
# -----------------------------------------


def loading_bar(name, delay):
    """
    Show a loading bar on the screen during a a certain amount of time
    Make the user understand the software is doing/waiting for something
    :param name: Text to be shown on the left of the loading bar
    :param delay: Amount of time the system is waiting in seconds
    :return: nothing
    """
    bar = IncrementalBar(name, max=(2 * delay), suffix='%(elapsed)s/' + str(delay) + ' seconds')
    for i in range(2 * delay):
        time.sleep(0.5)
        bar.next()
    bar.finish()
    return


def pump_start():
    """
    It put tension on the GPIO number 27 to turn on the air pump relay
    :return: nothing
    """
    GPIO.output(pump_gpio, GPIO.HIGH)


def pump_stop():
    """
    It remove tension from the GPIO number 27 to turn off the air pump relay
    :return: nothing
    """
    GPIO.output(pump_gpio, GPIO.LOW)


def append_data_to_csv(*data_to_write):
    """
    Store all the arguments given in the -data.csv file
    All arguments will be separated by a coma in the csv file
    :param data_to_write: unlimited amount of arguments
    :return: nothing
    """
    to_write = []
    for i in data_to_write:
        if isinstance(i, list):
            to_write += i
        else:
            to_write += [i]

    # print("To write in SD is:", to_write)

    # to_write = [*data_to_write]  # Concatenate all the arguments in one list

    with open(csv_file, mode='a', newline='') as data_file:
        writer = csv.writer(data_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(to_write)
    data_file.close()  # close file after use(general safety practice)


def wait_timestamp(starting_time):
    """
    Wait that the sampling period is passed out to start the next measurement
    :param starting_time: time at which the measurement has started ('time.time()' required)
    :return: Nothing, function will stop when the next measurement can start
    """
    finishing_time = time.time()
    next_launching = starting_time + sampling_period  # time at which the next sample should start

    to_wait = next_launching - finishing_time

    # if the sampling process took more time than the expected sampling period
    if finishing_time >= next_launching:
        logger.error("Sampling cycle took " + str(round((finishing_time - next_launching), 1)) +
                     "seconds more than expected (" + str(round(sampling_period, 0)) + " seconds)")
        print("Starting new sample...")
        return  # function stop and new sample start

    while time.time() < next_launching:
        now = time.time()
        for _ in range(10):
            now = time.time()
            if now < next_launching:
                time.sleep(.1)
            else:
                break
        to_wait = round(next_launching - now, 0)  # amount of time system will wait
        print("Waiting before next measurement:", int(to_wait) - 1, "seconds (sampling time is set on",
              sampling_period,
              "seconds)  ", end="\r")
        if not now < next_launching:
            break

    # Delete the waiting countdown and skip a line
    # Print a whole blank to remove completely the previous line ("Waiting before next measurement...")
    print("                                                                                    ")
    print("Starting new sample...")
    return  # function stop and new sample start


# --------------------------------------------
# MAIN CODE
# --------------------------------------------

# INITIATE CSV FILE

# Create the file to store the data if it doesn't exist
csv_file = directory_path + "/" + str(project_name) + "-data.csv"

to_write = []
to_write += ["Date/Time"]

# for the OPC-N3 sensor:
to_write += ["PM 1 (μg/m³)", "PM 2.5 (μg/m³)", "PM 10 (μg/m³)",
             "Temperature OPC (°C)", "Relative Humidity OPC (%RH)",
             "sampling time OPC (sec)", "sample flow rate OPC (ml/s)",
             "bin 0", "bin 1", "bin 2", "bin 3", "bin 4", "bin 5",
             "bin 6", "bin 7", "bin 8", "bin 9", "bin 10", "bin 11",
             "bin 12", "bin 13", "bin 14", "bin 15", "bin 16", "bin 17",
             "bin 18", "bin 19", "bin 20", "bin 21", "bin 22", "bin 23",
             "bin 1 MToF", "bin 3 MToF", "bin 5 MToF", "bin 7 MToF",
             "reject count glitch", "reject count long TOF", "reject count ratio",
             "reject count out of range", "fan revolution count", "laser status"]

# for the flow sensor:
to_write += ["flow (sccm)", "flow (slm)", "flow (slh)", "flow slm min", "flow slm max"]

# for the AFE:
to_write += ["temperature (°C)", "temperature (mV)",
             "temperature min", "temperature max",
             "NO2 (ppb)", "NO2 main (mV)", "NO2 aux (mV)",
             "NO2 main min", "NO2 main max", "NO2 aux min", "NO2 aux max",
             "OX (ppb)", "OX main (mV)", "OX aux (mV)",
             "OX main min", "OX main max", "OX aux min", "OX aux max",
             "SO2 (ppb)", "SO2 main (mV)", "SO2 aux (mV)",
             "SO2 main min", "SO2 main max", "SO2 aux min", "SO2 aux max",
             "CO2 (ppb)", "CO2 main (mV)", "CO2 aux (mV)"
             "CO main min", "CO main max", "CO aux min", "CO aux max"]

# for the GPS:
to_write += ["Date and time (UTC)",
             "GPS fix date and time (UTC)", "latitude", "longitude", "SOG (kts)", "COG",
             "horizontal dilution of precision", "accuracy",
             "altitude (m)", "WGS84 correction (m)", "fix type", "GPS status"]

# for the CO2 sensor:
to_write += ["Relative Humidity (%RH)", "Temperature (°C)", "Pressure (hPa)",
             "CO2 average (ppm)", "CO2 instant (ppm)"]

db_header = []
db_data_type = []

db_header += ["Date_Time"]
db_data_type += ["DATETIME"]

# for the OPC-N3 sensor:
db_header += ["PM_1", "PM_2_5", "PM_10",
             "Temperature_OPC", "Relative_Humidity_OPC",
             "sampling_time_OPC", "sample_flow_rate_OPC",
             "bin_0", "bin_1", "bin_2", "bin_3", "bin_4", "bin_5",
             "bin_6", "bin_7", "bin_8", "bin_9", "bin_10", "bin_11",
             "bin_12", "bin_13", "bin_14", "bin_15", "bin_16", "bin_17",
             "bin_18", "bin_19", "bin_20", "bin_21", "bin_22", "bin_23",
             "bin_1_MToF", "bin_3_MToF", "bin_5_MToF", "bin_7_MToF",
             "reject_count_glitch", "reject_count_long_TOF", "reject_count_ratio",
             "reject_count_out_of_range", "fan_revolution_count", "laser_status"]
db_data_type += ["FLOAT"] * 41

# for the flow sensor:
db_header += ["flow_sccm", "flow_slm", "flow_slh", "flow_slm_min", "flox_slm_max"]
db_data_type += ["FLOAT"] * (3+2)

# for the AFE:
db_header += ["temperature_C", "temperature_mV",
             "temperature_min", "temperature_max",
             "NO2_ppb", "NO2_main_mV", "NO2_aux_mV",
             "NO2_main_min", "NO2_main_max", "NO2_aux_min", "NO2_aux_max",
             "OX_ppb", "OX_main_mV", "OX_aux_mV",
             "OX_main_min", "OX_main_max", "OX_aux_min", "OX_aux_max",
             "SO2_ppb", "SO2_main_mV", "SO2_aux_mV",
             "SO2_main_min", "SO2_main_max", "SO2_aux_min", "SO2_aux_max",
             "CO2_ppb", "CO2_main_mV", "CO2_aux_mV",
             "CO_main_min", "CO_main_max", "CO_aux_min", "CO_aux_max"]
db_data_type += ["FLOAT"] * (32)

# for the GPS:
db_header += ["Date_time_UTC",
             "GPS_fix_date_and_time_UTC", "latitude", "longitude", "SOG", "COG",
             "horizontal_dilution", "accuracy",
             "altitude", "WGS84_correction", "fix_type", "GPS_status"]
db_data_type += ["TEXT"] * 12

# for the CO2 sensor:
db_header += ["Relative_Humidity", "Temperature", "Pressure",
             "CO2_average", "CO2_instant"]
db_data_type += ["FLOAT"] * 5



if mysql:
    db.create_new_table(db_header, db_data_type)

if not os.path.isfile(csv_file):  # if the file doesn't exist
    os.mknod(csv_file)  # create the file
    print("Initializing data file", csv_file)
    # Write a first line to the file, this will be the column headers
    append_data_to_csv(to_write)

else:
    logger.info("'" + str(csv_file) + "' already exist")
    logger.info("Appending data to this file")

print("\n########### STARTING SENSORS ############")

# Keep a trace of the day and time at which the system has started

now = datetime.now()  # get time
logger.info("Starting of Seacanairy on the " + str(now.strftime("%d/%m/%Y at %H:%M:%S")))  # delete time decimals

# WARN USER IF A SENSOR IS DEACTIVATED

if not CO2_activation:
    logger.warning("CO2 sensor has been disabled by the user in 'seacanairy_settings.yaml'")

if not OPCN3_activation:
    logger.warning("OPC-N3 sensor has been disabled by the user in 'seacanairy_settings.yaml'")

if not GPS_activation:
    logger.warning("GPS has been disabled by the user in 'seacanairy_settings.yaml'")

if not air_pump_activation:
    logger.warning("Air pump has been disabled by the user in 'seacanairy_settings.yaml'")

if not mysql:
    logger.warning("Database upload has been disabled by the user in 'seacanairy_settings.yaml'")

# SET INTERNAL CO2 SENSOR TIMESTAMP

if CO2_activation:
    # Read the internal timestamp of the CO2 sensor, and change the value if necessary
    if CO2_sampling_period != CO2.internal_timestamp():  # if internal timestamp is not the good one... change it
        CO2.internal_timestamp(CO2_sampling_period)
        # nothing in the brackets = read, timestamp in the brackets = write
else:  # if user don't want to use the sensor...
    try:
        # set the internal timestamp to the longest period available to reduce its wear...
        CO2.internal_timestamp(3600)
        logger.info("CO2 sensor disabled, internal timestamp set on 3600 seconds to reduce its wear")
    except:
        # It means that CO2 sensor is maybe not plugged in...
        pass  # nothing to do if it fails

if OPCN3_activation:
    # Initialize OPCN3 SPI, avoid loosing first communication
    OPCN3.initialization_SPI()
    time.sleep(.2)
    # Set the desired OPC fan speed
    OPCN3.set_fan_speed(OPC_fan_speed)

if CO2_activation:
    # Ask the CO2 sensor to take a new sample
    print("Synchronizing CO2 sensor with Seacanairy sampling period")
    CO2.trigger_measurement(True)  # True == force

print("\n############## SAMPLING! ###############")

# LOOP which will run continuously
try:  # necessary to stop the pump when the program is stopped
    while True:
        # Get the time in second at which the measurement start
        start = time.time()  # return the time expressed in second since the python date reference
        # a bit the same as 'millis()' on Arduino
        # easier to work with than with a complex string datetime format and a timedelta function...

        # Get date and time to store in the Excel file
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')  # remove the decimals and change the date order

        if air_pump_activation:
            # if OPCN3_activation:
            #     if OPCN3.fan_on():
            #         OPCN3.laser_on()
            #         OPCN3_fan_started = 1
            #         time.sleep(1)
            pump_start()  # start the pump
            print("Air pump is on")
            time.sleep(.5)  # let time to the pump to start and flow to happen

        to_write = [now]  # create the list in which data to store will be saved
        # [a, b, c] + [d, e, f] = [a, b, c, d, e, f]

        if CO2_activation:
            print("Triggering CO2 measurement")
            CO2.trigger_measurement()

        if flow_sensor_activation:
            # Start Flow Rate thread
            flow_period = 1
            flow_number_measurements = 10
            flow_delay = int(round(air_pump_minimum_running_time/2 - flow_period, 0))

            flow_thread = threading.Thread(target=flow.start_averaged_measurement,
                                           args=[flow_period, flow_number_measurements, flow_delay], daemon=True)
            print("Reading flow rate during", flow_period, "seconds after", flow_delay,
                  "seconds delay...")
            flow_thread.start()

        if OPCN3_activation:

            # Get OPC-N3 sensor data (see 'OPCN3.py')
            print("********************* OPC-N3 *********************")
            OPC_data = OPCN3.get_data(OPC_flushing_time, OPC_sampling_time, start_fan_laser=True)
            to_write += [OPC_data["PM 1"], OPC_data["PM 2.5"], OPC_data["PM 10"],
                         OPC_data["temperature"], OPC_data["relative humidity"],
                         OPC_data["sampling time"], OPC_data["sample flow rate"],
                         OPC_data["bin 0"], OPC_data["bin 1"], OPC_data["bin 2"], OPC_data["bin 3"],
                         OPC_data["bin 4"], OPC_data["bin 5"], OPC_data["bin 6"], OPC_data["bin 7"],
                         OPC_data["bin 8"], OPC_data["bin 9"], OPC_data["bin 10"], OPC_data["bin 11"],
                         OPC_data["bin 12"], OPC_data["bin 13"], OPC_data["bin 14"], OPC_data["bin 15"],
                         OPC_data["bin 16"], OPC_data["bin 17"], OPC_data["bin 18"], OPC_data["bin 19"],
                         OPC_data["bin 20"], OPC_data["bin 21"], OPC_data["bin 22"], OPC_data["bin 23"],
                         OPC_data["bin 1 MToF"], OPC_data["bin 3 MToF"],
                         OPC_data["bin 5 MToF"], OPC_data["bin 7 MToF"],
                         OPC_data["reject count glitch"],
                         OPC_data["reject count long TOF"], OPC_data["reject count ratio"],
                         OPC_data["reject count out of range"],
                         OPC_data["fan revolution count"], OPC_data["laser status"]]

        else:
            to_write += ["-"] * (41)

        if air_pump_activation:
            print("**************** AIR FLOW ****************")
            print("Reading flow rate...", end='\r')
            flow_thread.join()  # wait the end of the thread
            print("                      ", end='\r')
            flow_OPC_data = flow.get_averaged_measurement()
            to_write += [flow_OPC_data['average flow [sccm]'], flow_OPC_data['average flow [slm]'],
                         flow_OPC_data['average flow [slh]'],
                         flow_OPC_data['slm min'], flow_OPC_data['slm max']]

        else:
            to_write += ["-"] * (3+2)

        if air_pump_activation:
            while start + air_pump_minimum_running_time > time.time():
                #  wait until expected flushing time is reached
                print("Flushing air inside piping...", end='\r')
                time.sleep(.25)
            pump_stop()
            print("Air pump is off (pump has run", round(time.time() - start, 0), "seconds)")

        if AFE_activation:
            # Get OPC-N3 sensor data (see 'AFE.py')
            print("****************** AFE BOARD ********************")
            AFE_data = AFE.start_averaged_data(AFE_readings_averaged, display=True)
            to_write += [AFE_data["temperature"], AFE_data["temperature raw"],
                         AFE_data["temperature min"], AFE_data["temperature max"],
                         AFE_data["NO2 ppb"], AFE_data["NO2 main"], AFE_data["NO2 aux"],
                         AFE_data["NO2 main min"], AFE_data["NO2 main max"], AFE_data["NO2 aux min"], AFE_data["NO2 aux max"],
                         AFE_data["OX ppb"], AFE_data["OX main"], AFE_data["OX aux"],
                         AFE_data["OX main min"], AFE_data["OX main max"], AFE_data["OX aux min"],
                         AFE_data["OX aux max"],
                         AFE_data["SO2 ppb"], AFE_data["SO2 main"], AFE_data["SO2 aux"],
                         AFE_data["SO2 main min"], AFE_data["SO2 main max"], AFE_data["SO2 aux min"],
                         AFE_data["SO2 aux max"],
                         AFE_data["CO ppb"], AFE_data["CO main"], AFE_data["CO aux"],
                         AFE_data["CO main min"], AFE_data["CO main max"], AFE_data["CO aux min"],
                         AFE_data["CO aux max"],
                         ]

            print("")
            AFE.print_measurements(AFE_data)

        else:
            to_write += ["-"] * (4*3+4*4+4)

        if GPS_activation:
            # print("Wait switching from SPI to UART...", end='\r')
            # time.sleep(1)  # avoid too close communication between SPI of OPC and UART of GPS
            # Get GPS information
            print("********************** GPS **********************")
            GPS_data = GPS.get_position()
            to_write += [GPS_data["current time"], GPS_data["current date and time"],
                         GPS_data["latitude"], GPS_data["longitude"],
                         GPS_data["SOG"], GPS_data["COG"], GPS_data["horizontal precision"], GPS_data["accuracy"],
                         GPS_data["altitude"], GPS_data["WGS84 correction"], GPS_data["fix status"], GPS_data["status"]]

        else:
            to_write += ["-"] * 12

        if CO2_activation:
            # Get CO2 sensor data (see 'CO2.py')
            print("******************* CO2 SENSOR *******************")
            print("Waiting for sensor readiness...", end="\r")
            while time.time() < start + 16:
                time.sleep(.5)
            print("                                         ", end='\r')
            CO2_data = CO2.get_data()
            to_write += [CO2_data["relative humidity"], CO2_data["temperature"], CO2_data["pressure"],
                         CO2_data["average"], CO2_data["instant"]]

        else:
            to_write += ["-"] * 5

        # Store everything in the database
        if mysql:
            db.update(db_header, db_data_type, to_write)

        # Store everything in the csv file
        append_data_to_csv(to_write)

        # Time at which the sampling finishes
        finish = time.time()  # as previously, expressed in seconds since reference date

        # Calculate the amount of time the sampling process took, round to 0 to avoid decimals
        # int(...) to delete the remaining 0 behind the coma
        logger.info("Sampling finished in " + str(int(round(finish - start, 0))) + " seconds")

        # Wait that the sampling period is passed
        wait_timestamp(start)

except:  # stop everything before exiting the program
         logger.critical("Software has been stopped (" + str(sys.exc_info()) + ")")
         pump_stop()
         # OPCN3.laser_off()
         # OPCN3.fan_off()